package tasks;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.Random;


public abstract class HelperFunctions {

    /**
     * Klasse, die ein Bild repräsentiert.
     */
	
    public static class Image {
        /** Grauwerte des Bildes */
    	
        private int[] data = new int[0];

		/** Breite des Bildes */
        
        private int w = 0;
        /** Höhe des Bildes */
        
        private int h = 0;

        /** Headerdaten */
        
        private int[] header = new int[0];

        /**
         * Erzeugt ein leeres Bild.
         */
        
        Image() {
        }

        /**
         * Legt ein neues Bild-Objekt an und kopiert die
         * Daten des eingegebenen Bildes in das neue Objekt.
         * @param old ein Bild, das kopiert werden soll.
         */
        
        Image(final Image old) {
            w = old.w;
            h = old.h;
            data = new int[old.data.length];
            System.arraycopy(old.data, 0, data, 0, data.length);
            header = new int[old.header.length];
            System.arraycopy(old.header, 0, header, 0, header.length);
        }
        
        /**
         * Gibt die Grauwerte des Bildes zurück.
         * @return die Grauwerte des Bildes
         */
        
        public int[] getData() {
			return data;
		}

        /**
         * Setzt die Grauwerte des Bildes
         * @param data die Grauwerte des Bildes
         */
        
		public void setData(int[] data) {
			this.data = data;
		}

		/**
		 * Gibt die Höhe des Bildes zurück.
		 * @return die Höhe des Bildes 
		 */
		
		public int getWidth() {
			return w;
		}

		/**
		 * Setzt die Höhe des Bildes.
		 * @param w die Höhe des Bildes
		 */
		
		public void setWidth(int w) {
			this.w = w;
		}

		/**
		 * Gibt die Breite des Bildes zurück.
		 * @return die Breite des Bildes zurück
		 */
		
		public int getHeight() {
			return h;
		}

		/**
		 * Setzt die Höhe des Bildes.
		 * @param h die Höhe des Bildes.
		 */
		
		public void setHeight(int h) {
			this.h = h;
		}

		/**
		 * Gibt die Größe des Bildes zurück (Anzahl der Pixel)
		 * @return die Größe des Bildes
		 */
		
		public int getSize() {
			return data.length;
		}

		/**
		 * Gibt die Headerinformationen des Bildes zurück.
		 * @return die Headerinformationen des Bildes
		 */
		
		public int[] getHeader() {
			return header;
		}

		/**
		 * Setzt die Headerinformationen des Bildes.
		 * @param header die Headerinformationen des Bildes
		 */
		
		public void setHeader(int[] header) {
			this.header = header;
		}
        
    }

    /**
     * Einlesen eines Bildes im pgm-Format.
     * @param file Bilddatei
     * @return das Bild
     * @throws Exception wenn Fehler beim Einlesen des Bildes auftreten.
     */
    
    public static Image read(final File file) throws Exception {
        final int hwSize = 21;
        int[] heightWidth = new int[hwSize];
        int[] content;

        long fileLength = 0;
        int lfCounter = 0;
        boolean comment = false, abort = false;
        boolean heightWidthFound = false, heightEnd = false;
        final int asciiCR = 10;
        final int asciiLF = 13;
        final int asciiSPACE = 32;
        FileInputStream fis = null;
        Image result = new Image();

        fis = new FileInputStream(file);
        fileLength = file.length();

        // Read the file 
        content = new int[(int)fileLength];
        byte [] b = new byte[(int)fileLength];
        int r = fis.read(b);
        if (r < content.length) {
        	printf(tasks.Main.messages.getString("helperFunctions_errorEarlyEOF"));
        	fis.close();
        	return null;
        }
        // convert grayvalues to int
        for(int i = 0; i < fileLength; i++) {
        	content[i] = b[i] >= 0 ? (int)b[i] : 256 + (int)b[i];
        }
        fis.close();
        // detect header length
        int headerLength = 0;
        for (int a = 0; a < fileLength; a++) {
            // without comments, there are 3 LF in the header
            // therefore we need 3 loops max
            if (lfCounter <= 2) {
            	// decrease LF counter for every comment line
                if (content[a] == '#') {
                    lfCounter--;
                }
                // increase LF counter at LF
                if (content[a] == asciiCR || content[a] == asciiLF) {
                    lfCounter++;
                }
                headerLength++;
            }
        }

        result.header = new int[headerLength];
        // read the header
        for (int a = 0; a < headerLength; a++) {
            result.header[a] = content[a];
        }
        // get height and with
        // both values are located after the first (non-comment) LF
        int pos = 0;
        for (int i = 0, a = 0; i < headerLength - 1; i++) {
            if (!abort) {
                if (heightWidthFound) {
                    // read until the next LF 
                    if (result.header[i] != asciiCR || result.header[i] != asciiLF) {
                        heightWidth[a] = result.header[i];
                        a++;
                    } else {
                        abort = true;
                    }
                }
                // detect comment line 
                if (result.header[i] == '#') {
                    comment = true;
                }
                // detect comment line end
                if ((result.header[i] == asciiCR || result.header[i] == asciiLF)
                        && comment) {
                    comment = false;
                }
                // detect first non-comment LF, after which height and with are located
                if ((result.header[i] == asciiCR || result.header[i] == asciiLF)
                        && !comment) {
                    if (result.header[i + 1] != '#') {
                        heightWidthFound = true;
                    }
                }
            }
            pos = a;
        }
        // parse height and width
        String height = "", width = "";
        for (int i = 0; i <= pos + 1; i++) {
            // detect space between height and width
            if (heightWidth[i] == asciiSPACE) {
                heightEnd = true;
            }
            if (heightWidth[i] == asciiCR || heightWidth[i] == asciiLF) {
                break;
            }
            // width is before the space
            if (heightWidth[i] != asciiSPACE && !heightEnd) {
                width += (char)heightWidth[i];
            }
            // height is after the space
            if (heightWidth[i] != asciiSPACE && heightEnd) {
                height += (char)heightWidth[i];
            }
        }
        
        // read geay values
        result.h = Integer.parseInt(height);
        result.w = Integer.parseInt(width);

        int size = result.h * result.w;
        result.data = new int[size];
        for (int i = headerLength, h = 0; h < size; h++, i++) {
            result.data[h] = content[i];
        }
        return result;
    }


    /**
     * Schreibt das Bild im pgm-Format in eine Datei.
     * @param path Pfad und Dateiname für die Ausgabedatei
     * @param image das auszugebende Bild
     * @throws Exception wenn Fehler beim Schreiben des Bildes auftreten.
     */
    
    public static void write(final String path, final Image image) throws Exception {
        FileOutputStream fos = new FileOutputStream(path);

        if (fos == null) {
            printf(tasks.Main.messages.getString("helperFunctions_errorOpenFileForWrite"));
        } else {
            // write the header
            int headerLength = image.header.length;
            byte[] tmpArray = new byte[headerLength];
            for (int i = 0; i < headerLength; i++) {
                tmpArray[i] = (byte) ((char)(image.header[i]%256));
            }
            fos.write(tmpArray);

            // write the gray values
            int dataLength = image.data.length;
            tmpArray = new byte[dataLength];
            for (int i = 0; i < dataLength; i++) {
                tmpArray[i] = (byte) ((char)(image.data[i]%256));
            }
            fos.write(tmpArray);
            fos.close();
        }
    }

    /**
     * Einlesen des Schlüssels aus der vorgegebenen Datei
     * "key.txt"
     * @return der Schlüssel
     * @throws Exception wenn Fehler beim Einlesen auftreten.
     */
    
    public static int[] readKey() throws Exception {
        int fileLength = 0;
        FileInputStream fis = null;
        int[] result = null;
        final int asciiZero = 48;
        final int asciiNine = 58;

        File f = new File("key.txt");
        fis = new FileInputStream(f);
        fileLength = (int) f.length();

        if (fis == null) {
            printf(tasks.Main.messages.getString("helperFunctions_errorOpenFileForRead"));
            result = null;
        } else {
            int[] temp = new int[fileLength];

            byte b [] = new byte[fileLength];
            int r = fis.read(b);
            if (r < b.length) {
            	printf(tasks.Main.messages.getString("helperFunctions_errorEarlyEOFForKey"));
                return null;
            }
            int l = 0;
            for (int i = 0; i < fileLength; i++) {
                // read only numbers
                if (b[i] >= asciiZero && b[i] <= asciiNine) {
                    temp[l++] = ((char)b[i])-'0';
                }
            }
            result = new int[l];
            System.arraycopy(temp, 0, result, 0, l);

            fis.close();
        }
        return result;
    }


    /**
     * Funktion zum Auslesen der Nachricht aus der vorgegebenen Datei
     * "message.txt"
     * @return die Nachricht in Form einer ASCII-Zeichenkette
     * @throws Exception wenn Fehler beim Einlesen der Nachricht auftreten.
     */
    
    public static boolean[] readMessage() throws Exception {
        FileInputStream fis = null;
        boolean[] result = null;

        File f = new File("message.txt");
        fis = new FileInputStream(f);
        int fileLength = (int) f.length();

        if (fis == null) {
            printf(tasks.Main.messages.getString("helperFunctions_errorOpenFileForRead"));
            result = null;
        } else {
            byte b [] = new byte[fileLength];
            int r = fis.read(b);
            if (r < b.length) {
            	printf(tasks.Main.messages.getString("helperFunctions_errorEarlyEOFForMessage"));
                return null;
            }
            result = new boolean[fileLength*8];
            for(int i = 0; i < fileLength; i++) {
                for (int j = 0; j < 8; j++) {
                    result[(i * 8) + j] = ((b[i] >> j) % 2) == 1;
                }
            }
            fis.close();
        }
        return result;
    }


    /**
     * Funktion zum Erzeugen einer pseudozufälligen Binär-Nachricht. 
     * @param size gibt die Länge der Nachricht in Bit an.
     * @return eine zufällige Nachricht.
     */
    
   public static boolean[] getRnd(final int size) {

        boolean[] result = new boolean[size];
        // use a seed in order to get always the same message for comparisions 
        final long seed = 123456;
        Random rand = new Random(seed);

        for (int i = 0; i < size; i++) {
            result[i] = rand.nextInt(2) > 0;
        }
        return result;
    }

    /**
     * Gibt einen String auf der Standardausgabe aus.
     * @param s der auszugebende String.
     */
   
    private static void printf(final String s) {
        System.out.print(s);
    }

}
